package org.tsutils;



/**
 * To generate header: javah -classpath target/classes -jni -d src/main/c org.tsutils.RdtscTimestamper 
 * 
 * To run: java -cp target/classes -Djava.library.path=src/main/resources/linux org.tsutils.RdtscTimestamper
 * 
 * Based on Peter Lawrey code: https://github.com/peter-lawrey/Java-Thread-Affinity/blob/master/src/main/java/vanilla/java/clock/impl/JNIClock.java
 * 
 * @author Sergio Oliveira Jr.
 */
public class RdtscTimestamper extends AbstractTimestamper {
	
    private static final int FACTOR_BITS = 17;
    private static final int WARMUP = 4;
    private static final int COUNT = 4;
    private static final int FACTOR = 1000;
    
    private static boolean AVAILABLE = false;

    private final long start;
    private long factorValue;
    private final boolean debug;
    
    public RdtscTimestamper() {
    	this(false);
    }
    
    public RdtscTimestamper(boolean debug) {
    	super();
    	this.debug = debug;
    	if (!AVAILABLE) {
    		String os = System.getProperty("os.name").toLowerCase();
    		boolean isLinux = os.indexOf("nix") >= 0 || os.indexOf("nux") >= 0 || os.indexOf("aix") > 0;
    		if (isLinux) {
    			String filename = "lib" + RdtscTimestamper.class.getSimpleName() + ".so";
    			throw new IllegalStateException("Cannot find native libraries: " + filename);
    		} else {
    			throw new IllegalStateException("Cannot find native libraries!");
    		}
    	}
    	this.start = rdtsc();
    	this.factorValue = calcFrequency(FACTOR, WARMUP, COUNT);
    }
    
    private long calcFrequency(int factor, int warmup, int count) {
    	// first warmup...
    	for(int i = 0; i < warmup; i++) {
    		long freq = estimateFrequency(factor);
    		if (debug) {
    			System.out.println("Warmup frequency: " + freq);
    		}
    	}
    	long[] values = new long[count];
    	for(int i = 0; i < count; i++) {
    		values[i] = estimateFrequency(factor);
    		if (debug) {
    			System.out.println("Estimated frequency: " + values[i]);
    		}
    	}
    	
    	long finalValue = average(values);
    	if (debug) {
    		System.out.println("Average frequency: " + finalValue);
    	}
    	
    	return finalValue;
    }
    
    private long average(long[] values) {
    	long total = 0;
    	for(int i = 0; i < values.length; i++) {
    		total += values[i];
    	}
    	return total / values.length;
    }
	
    private final long tscToNano(final long tsc) {
        return (tsc * factorValue) >> FACTOR_BITS;
    }
    
	private final native long rdtsc();
	
	@Override
    public final long nanoTime() {
        return tscToNano(rdtsc() - start);
    }
	
	static {
		try {
			System.loadLibrary(RdtscTimestamper.class.getSimpleName());
			AVAILABLE = true;
		} catch(Throwable e) {
			// ignore and throw exception only when class is instantiated (see constructor)
		}
		
	}
	
	public static boolean isAvailable() {
		return AVAILABLE;
	}
	
    private long estimateFrequency(int factor) {
    	
    	final long start = System.nanoTime();
        long now;
        while ((now = System.nanoTime()) == start);
        long end = start + factor * 1000000L;
        final long start0 = rdtsc();
        while ((now = System.nanoTime()) < end);
        long end0 = rdtsc();
        end = now;

        return ((end - start) << FACTOR_BITS) / (end0 - start0) - 1;
    }
	
	public static void main(String[] args) {
		Timestamper ts = new RdtscTimestamper(true);
		System.out.println(ts.nanoTime());
		System.out.println(ts.now());
	}
}